"""

"""

# Version and library information
LIBRARY_NAME = "pywib"
LIBRARY_VERSION = "0.1.0"

# Event types for interaction tracking
class EventTypes:
    """ Event type constants for interaction tracking."""
    EVENT_ON_MOUSE_MOVE = 0
    """Event type for mouse move events."""
    EVENT_ON_CLICK = 1
    """Event type for mouse click events."""
    EVENT_ON_DOUBLE_CLICK = 2
    """Event type for mouse double click events."""
    EVENT_ON_MOUSE_DOWN = 3
    """Event type for mouse down events."""
    EVENT_ON_MOUSE_UP = 4
    """Event type for mouse up events."""
    EVENT_ON_WHEEL = 5
    """Event type for mouse wheel events, specificaly for wheel clicks."""
    EVENT_CONTEXT_MENU = 6
    """Event type for context menu events."""
    EVENT_ON_TOUCH_MOVE = 7
    """Event type for touch move events, specific for mobile and tablet devices."""
    EVENT_WINDOW_SCROLL = 11
    """Event type for window scroll events."""
    EVENT_WINDOW_RESIZE = 12
    """Event type for window resize events."""
    EVENT_KEY_DOWN = 13
    """Event type for key down events."""
    EVENT_KEY_PRESS = 14
    """Event type for key press events."""
    EVENT_KEY_UP = 15
    """Event type for key up events."""
    EVENT_FOCUS = 16
    """Event type for focus events."""
    EVENT_BLUR = 17
    """Event type for blur events."""
    EVENT_ON_CHANGE_SELECTION_OBJECT = 18
    """Event type for change selection events."""
    EVENT_ON_CLICK_SELECTION_OBJECT = 19
    """Event type for click selection events."""
    EVENT_INIT_TRACKING = 100
    """Custom event type for initializing tracking."""
    EVENT_TRACKING_END = 200
    """Custom event type for ending tracking."""

class ComponentTypes:
    """ Component type constants for UI elements."""
    COMPONENT_TEXT_FIELD = 1
    COMPONENT_COMBOBOX = 2
    COMPONENT_OPTION = 3
    COMPONENT_RADIO_BUTTON = 4
    COMPONENT_CHECK_BOX = 5

class ColumnNames:
    """ Standard column names for DataFrame operations."""
    SESSION_ID = 'sessionId'
    SCENE_ID = 'sceneId'
    EVENT_TYPE = 'eventType'
    ELEMENT_ID = 'elementId'
    TIME_STAMP = 'timeStamp'
    X = 'x'
    Y = 'y'
    KEY_VALUE_EVENT = 'keyValueEvent'
    KEY_CODE_EVENT = 'keyCodeEvent'
    SOURCE_SESSION_ID = 'sourceSessionId'
    DT = 'dt'
    DX = 'dx'
    DY = 'dy'
    VELOCITY= 'velocity'
    ACCELERATION = 'acceleration'
    JERKINESS = 'jerkiness'
    AUC_RATIO = 'auc_ratio'

class KeyCodeEvents:
    """ Key code event constants for keyboard interactions."""
    KEY_CODE_BACKSPACE = 8
    """Key code for the Backspace key."""
    KEY_CODE_DELETE = 46
    """Key code for the Delete key."""


import pandas as pd
from ..constants import EventTypes, ColumnNames
from ..utils.validation import validate_dataframe, validate_dataframe_keyboard

def extract_trace(dt: pd.DataFrame) -> list:
    """
    Extracts trace from the DataFrame.
    Each trace is considered as a sequence of consecutive ON_MOUSE_MOVE events
    between two non-move events.
    Returns a list of DataFrames, each corresponding to a trace.
    """
    validate_dataframe(dt)
    dt = dt.sort_values(by=ColumnNames.TIME_STAMP).reset_index(drop=True)
    is_move = (dt[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_MOUSE_MOVE) | (dt[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_TOUCH_MOVE)
    group_id = (~is_move).cumsum()
    traces = []
    for _, group in dt[is_move].groupby(group_id[is_move]):
        if len(group) > 1:
            traces.append(group)
    return traces

def extract_traces_by_session(dt: pd.DataFrame) -> dict:
    """
    Extracts traces from the DataFrame, grouped by (sessionId, sceneId).
    Each trace is considered as a sequence of consecutive ON_MOUSE_MOVE events
    between two non-move events.
    Parameters:
        dt (pd.DataFrame): DataFrame containing 'sessionId', 'sceneId', 'eventType', and 'timeStamp' columns.
    Returns:
        dict: a dictionary with keys as (sessionId) and values as lists of DataFrames.
    """
    validate_dataframe(dt)
    dt = dt.sort_values(by=ColumnNames.TIME_STAMP).reset_index(drop=True)
    traces_by_session = {}
    for session_id, group in dt.groupby(ColumnNames.SESSION_ID):
        is_move = group[ColumnNames.EVENT_TYPE].isin([EventTypes.EVENT_ON_MOUSE_MOVE, EventTypes.EVENT_ON_TOUCH_MOVE])
        group_id = (~is_move).cumsum()
        traces = []
        for _, sub_group in group[is_move].groupby(group_id[is_move]):
            if len(sub_group) > 1:
                traces.append(sub_group)
        traces_by_session[session_id] = traces
    return traces_by_session

def extract_keystroke_traces_by_session(dt: pd.DataFrame) -> dict[str, list[pd.DataFrame]]:
    """
    Extracts keystroke traces from the DataFrame, grouped by (sessionId, sceneId).
    Each keystroke trace is considered as a sequence of consecutive key events
    between two non-keyboard events.
    Parameters:
        dt (pd.DataFrame): DataFrame containing 'sessionId', 'sceneId', 'eventType', and 'timeStamp' columns.
    Returns:
        dict: a dictionary with keys as (sessionId) and values as lists of DataFrames.
    """
    validate_dataframe_keyboard(dt)
    dt = dt.sort_values(by=ColumnNames.TIME_STAMP).reset_index(drop=True)
    keystroke_traces_by_session = {}
    for session_id, group in dt.groupby(ColumnNames.SESSION_ID):
        is_keydown = group[ColumnNames.EVENT_TYPE].isin([EventTypes.EVENT_KEY_UP, EventTypes.EVENT_KEY_DOWN, EventTypes.EVENT_KEY_PRESS])
        group_id = (~is_keydown).cumsum()
        keystroke_traces = []
        for _, sub_group in group[is_keydown].groupby(group_id[is_keydown]):
            if len(sub_group) > 0:
                keystroke_traces.append(sub_group)
        keystroke_traces_by_session[session_id] = keystroke_traces
    return keystroke_traces_by_session
    
def extract_mouse_click_traces_by_ession(dt: pd.DataFrame) -> dict:
    """
    Extracts those traces with event movements that end with ON_MOUSE_CLICK or ON_TOUCH_TAP events,
    grouped by sessionId.
    """
    validate_dataframe(dt)
    dt = dt.sort_values(by=ColumnNames.TIME_STAMP).reset_index(drop=True)
    click_traces_by_session = {}
    for session_id, group in dt.groupby(ColumnNames.SESSION_ID):
        is_move = (group[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_MOUSE_MOVE) | (group[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_TOUCH_MOVE)
        group_id = (~is_move).cumsum()
        click_traces = []
        for _, sub_group in group[is_move].groupby(group_id[is_move]):
            if len(sub_group) > 1:
                last_index = sub_group.index[-1]
                # get the integer position of last_index within `group`'s index to avoid assuming contiguous indices
                try:
                    pos = group.index.get_loc(last_index)
                except KeyError:
                    # if for some reason the index is not found, skip this sub_group
                    continue
                if pos + 1 < len(group):
                    next_event = group.iloc[pos + 1][ColumnNames.EVENT_TYPE]
                    if next_event in [EventTypes.EVENT_ON_CLICK, EventTypes.EVENT_ON_MOUSE_DOWN, EventTypes.EVENT_ON_MOUSE_UP]:
                        click_traces.append(sub_group)
        click_traces_by_session[session_id] = click_traces
    return click_traces_by_session

def extract_mouse_click_traces_by_session_with_intial_pause(dt: pd.DataFrame, pause_threshold: float = 200) -> dict:
    dt = dt.sort_values(by=ColumnNames.TIME_STAMP).reset_index(drop=True)
    click_traces_by_session = {}
    for session_id, group in dt.groupby(ColumnNames.SESSION_ID):
        is_move = (group[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_MOUSE_MOVE) | (group[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_TOUCH_MOVE)
        group_id = (~is_move).cumsum()
        click_traces = []
        for _, sub_group in group[is_move].groupby(group_id[is_move]):
            if len(sub_group) > 1:
                last_index = sub_group.index[-1]
                # get the integer position of last_index within `group`'s index to avoid assuming contiguous indices
                try:
                    pos = group.index.get_loc(last_index)
                except KeyError:
                    # if for some reason the index is not found, skip this sub_group
                    continue
                if pos + 1 < len(group):
                    next_event = group.iloc[pos + 1][ColumnNames.EVENT_TYPE]
                    if next_event in [EventTypes.EVENT_ON_CLICK, EventTypes.EVENT_ON_MOUSE_DOWN, EventTypes.EVENT_ON_MOUSE_UP]:
                        sub_group = sub_group.sort_values(by=ColumnNames.TIME_STAMP).reset_index(drop=True)
                        sub_group[ColumnNames.DT] = sub_group[ColumnNames.TIME_STAMP].diff().fillna(0)
                        pauses = sub_group[sub_group[ColumnNames.DT] > pause_threshold]
                        # Now check if there is a pause at the beginning and no pauses in between
                        if not pauses.empty and pauses.index[0] == 1 and pauses.shape[0] == 1:
                            click_traces.append(sub_group)

        click_traces_by_session[session_id] = click_traces
    return click_traces_by_session


"""
Analysis module for HCI Web Interaction Analyzer

This module provides methods to analyze interaction data from DataFrames.
"""

import pandas as pd
import numpy as np
from pywib.utils.segmentation import extract_keystroke_traces_by_session
from pywib.utils.validation import validate_dataframe_keyboard
from pywib.constants import EventTypes, ColumnNames, KeyCodeEvents
from pywib.utils.keyboard import (backspace_usage_df, backspace_usage_traces, typing_durations_df, typing_durations_traces, typing_speed_df, typing_speed_traces)

def typing_durations(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None, per_traces: bool = True) -> list:
    """
    Calculate the durations of individual keystrokes.

    Parameters:
        df (pd.DataFrame): DataFrame containing interaction data with 'event_type', 'timestamp', and 'key' columns.
        traces (dict[str, list[pd.DataFrame]]): optional Pre-extracted keystroke traces by session.
        per_traces (bool): optional Whether to calculate durations per trace. Default is True.
    Returns:
        list: List of keystroke durations in milliseconds.
    """
    if traces is None and per_traces:
        traces = extract_keystroke_traces_by_session(df)
        return typing_durations_traces(traces, False)

    if not per_traces:
        return typing_durations_df(df)

    return typing_durations_traces(traces)

def typing_speed(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None, per_traces : bool = True) -> dict[list[float]] | float:
    """
    Calculate the average typing speed in characters per minute (CPM).

    Parameters:
        df (pd.DataFrame): DataFrame containing interaction data with 'event_type', 'timestamp', and 'key' columns.
        traces (dict[str, list[pd.DataFrame]]): optional Pre-extracted keystroke traces by session.
        per_traces (bool): optional Whether to calculate speed per trace. Default is True. if False, mind that the df must only contain typing data in order to obtain the correct CPM calculation.

    Returns:
        dict (dict[list[float]] | float) : A dictionary with session IDs as keys and lists of typing speeds (CPM) per trace as values, or a float representing the typing speed if per_traces is False.
    """

    if per_traces and traces is None:
        traces = extract_keystroke_traces_by_session(df)
        return typing_speed_traces(traces, False)

    elif not per_traces:
        return typing_speed_df(df)

    return typing_speed_traces(traces)

def typing_speed_metrics(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None) -> dict:
    """
    Calculate typing speed metrics including average CPM, total characters typed, and total time spent typing.
    
    This metrics include average typing speed (average_typing_speed) in CPM, total characters typed (total_characters), and total time spent typing (total_time_seconds) in seconds.
    
    The average keydown to keyup duration (avg_keydown_to_keyup_duration) is the average duration of a keystroke (from keydown to keyup) in milliseconds.

    Parameters:
        df : pd.DataFrame DataFrame containing interaction data with 'event_type', 'timestamp', and 'key' columns.
        traces : dict[str, list[pd.DataFrame]], optional Pre-extracted keystroke traces by session.
        per_trace : bool, optional Whether to calculate metrics per trace. Default is True.
    Returns:
        dict: A dictionary with session IDs as keys and their corresponding typing speed metrics as values.
        """
    if traces is None:
        traces = extract_keystroke_traces_by_session(df)
    
    session_speeds = typing_speed(None, traces=traces, per_traces=True)

    metrics_by_session = {}
    for session_id, speeds in session_speeds.items():
        if speeds:
            avg_speed = np.mean(speeds)
            total_chars = sum(session_traces[session_traces[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_KEY_UP][ColumnNames.EVENT_TYPE].count() for session_traces in traces[session_id])
            total_time = sum((session_traces[ColumnNames.TIME_STAMP].diff().fillna(0).sum() / 1000.0) for session_traces in traces[session_id])
            avg_keydown_to_keyup_duration = np.mean([
                (trace[trace[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_KEY_UP][ColumnNames.TIME_STAMP].values - trace[trace[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_KEY_DOWN][ColumnNames.TIME_STAMP].values).mean()
                for trace in traces[session_id]
            ])
            metrics_by_session[session_id] = {
                "average_typing_speed": avg_speed,
                "total_characters": total_chars,
                "total_time_seconds": total_time,
                "avg_keydown_to_keyup_duration": avg_keydown_to_keyup_duration  # TODO Review
            }
    return metrics_by_session


def backspace_usage(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None, per_trace: bool = True) -> dict:
    """
    Calculate the backspace usage rate (backspaces per 100 characters typed) for each session.

    Parameters:
        df (pd.DataFrame): DataFrame containing interaction data with 'event_type', 'timestamp', and 'key' columns.
        traces (dict[str, list[pd.DataFrame]]): optional Pre-extracted keystroke traces by session.
        per_trace (bool): optional Whether to calculate usage per trace. Default is True.
    Returns:
        dict: A dictionary with session IDs as keys and their corresponding backspace counts as values.
    """
    if traces is None and per_trace:
        validate_dataframe_keyboard(df)
        traces = extract_keystroke_traces_by_session(df)
        return backspace_usage_traces(traces, False)
    if not per_trace:
        validate_dataframe_keyboard(df)
        return backspace_usage_df(df)
    
    return backspace_usage_traces(traces)


import pandas as pd
from collections import defaultdict

def import_pyModule():
    """
    Necessary to import the pywib package when running tests directly from the test/ folder.
    """
    import sys
    import os

    # If running tests from the repo (package not installed), add src/ to sys.path
    REPO_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), '..'))
    SRC_PATH = os.path.join(REPO_ROOT, 'src')
    if SRC_PATH not in sys.path:
        sys.path.insert(0, SRC_PATH)

def process_csv(file_path):
    """
    Reads a semicolon-separated CSV file and processes it into matrices grouped by sessionId and sceneId.
    Args:
        file_path (str): Path to the CSV file.
    """
    # Read the CSV file with semicolon separator
    df = pd.read_csv(file_path, encoding='utf-8', sep=',')
    
    # Dictionary to store matrices for each sessionId and sceneId
    matrices = defaultdict(list)

    # Process each row in the DataFrame
    for _, row in df.iterrows():
        session_id = row['sessionId']
        
        # Create a key combining sessionId and sceneId
        key = (session_id)
        
        # Extract the relevant values and append to the matrix
        matrices[key].append([
            row['eventType'],
            row['timeStamp'],
            row['x'],
            row['y'],
            row['keyValueEvent'],
            row['keyCodeEvent'],
        ])
    
    all_sessions = []
    for (session_id), matrix in matrices.items():
        df = pd.DataFrame(matrix, columns=[
            'eventType', 'timeStamp', 'x', 'y', 'keyValueEvent', 'keyCodeEvent' ])
        df['sessionId'] = session_id
        all_sessions.append(df)

    df_all_sessions = pd.concat(all_sessions, ignore_index=True)
    df_all_sessions['timeStamp'] = df_all_sessions['timeStamp'].astype(str).str.replace(',', '', regex=False)
    df_all_sessions['timeStamp'] = pd.to_numeric(df_all_sessions['timeStamp'], errors='coerce')
    return df_all_sessions

def csv_to_df_no_checks(file_path):
    """
    Reads a semicolon-separated CSV file into a DataFrame without any validation checks.
    Args:
        file_path (str): Path to the CSV file.
    """
    return pd.read_csv(file_path, encoding='utf-8', sep=',')



import pandas as pd
from ..constants import ColumnNames
from ..utils.validation import validate_dataframe

def compute_space_time_diff(df: pd.DataFrame) -> pd.DataFrame:
    """
    Compute space and time differences (dx, dy, dt) for the given DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'x', 'y', and 'timeStamp' columns.
    
    Returns:
        pd.DataFrame: DataFrame with additional 'dx', 'dy', and 'dt' columns.
    """
    if(ColumnNames.X not in df.columns or
       ColumnNames.Y not in df.columns or
       ColumnNames.TIME_STAMP not in df.columns or
       ColumnNames.SESSION_ID not in df.columns):
        raise ValueError(f"DataFrame must contain '{ColumnNames.X}', '{ColumnNames.Y}', '{ColumnNames.TIME_STAMP}', and '{ColumnNames.SESSION_ID}' columns.")

    df = df.copy()
    df.sort_values(by=[ColumnNames.TIME_STAMP], inplace=True)
    df[ColumnNames.TIME_STAMP] = pd.to_numeric(df[ColumnNames.TIME_STAMP], errors='coerce')
    df['dt'] = df.groupby([ColumnNames.SESSION_ID])[ColumnNames.TIME_STAMP].diff().fillna(0)
    df['dx'] = df.groupby([ColumnNames.SESSION_ID])[ColumnNames.X].diff().fillna(0)
    df['dy'] = df.groupby([ColumnNames.SESSION_ID])[ColumnNames.Y].diff().fillna(0)
    return df

def compute_metrics_from_traces(
    df: pd.DataFrame,
    traces: dict[str, list[pd.DataFrame]] | None,
    column_name: str,
    compute_traces_fn,
    preprocess_fn=None
) -> dict:
    """
    Compute basic statistical metrics (mean, max, min) for a specific column across sessions.

    This function serves as a generic helper for computing metrics such as velocity or acceleration
    over session-based trace data. It can automatically compute traces if they are not provided and
    allows for custom preprocessing (e.g., filtering out zero values).

    Parameters:
        df (pd.DataFrame): 
            A DataFrame containing the necessary data. Must include the specified column 
            or enough information to compute it via `compute_traces_fn`.

        traces (dict[str, list[pd.DataFrame]] | None): 
            Optional dictionary containing session traces. 
            Each key corresponds to a sessionId, and its value is a list of DataFrames 
            representing that session's traces. If None, traces will be computed using 
            `compute_traces_fn(df)`.

        column_name (str): 
            The name of the column from which to compute metrics (e.g., "velocity", "acceleration").

        compute_traces_fn (Callable): 
            A function that, given a DataFrame, computes and returns the corresponding traces dictionary.

        preprocess_fn (Callable | None): 
            Optional function applied to the concatenated column values before computing statistics. 
            Typically used to filter out zero or invalid values.

    Returns:
        dict:
            A dictionary where keys are sessionIds and values are dictionaries containing:
            - 'mean': Mean value.
            - 'max': Maximum value.
            - 'min': Minimum value.
    """
    if (traces is None) or (column_name not in df.columns):
        validate_dataframe(df)
        traces = compute_traces_fn(df, per_traces=True)

    metrics = {}
    for session_id, session_traces in traces.items():
        # TODO error on empty session_traces?
        values = pd.concat([trace[column_name] for trace in session_traces])
        if preprocess_fn:
            values = preprocess_fn(values)
        metrics[session_id] = {
            'mean': values.mean(),
            'max': values.max(),
            'min': values.min()
        }

    return metrics


import pandas as pd

required_columns = [
    "id", "eventType", "timeStamp", "x", "y"
]

keyboard_columns = [
   "keyValueEvent", "keyCodeEvent"
]

def validate_dataframe(df: pd.DataFrame):

    for col in required_columns:
        if col not in df.columns:
            raise ValueError(f"Missing required column: {col}")
        
def validate_dataframe_keyboard(df: pd.DataFrame):

    columns_to_check = required_columns + keyboard_columns

    for col in columns_to_check:
        if col not in df.columns:
            raise ValueError(f"Missing required column: {col}")


import pandas as pd
from ..utils.validation import validate_dataframe
from ..utils.segmentation import extract_traces_by_session
from ..utils.utils import compute_space_time_diff
from ..constants import ColumnNames

def execution_time(df: pd.DataFrame) -> dict:
    """
    Calculate the total execution time of a session in milliseconds, without taking pauses into account.
    This is the same as the total time from the first to the last event registered for the session.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'timeStamp' column.
    Returns:
        dict: Total execution time in milliseconds for each session.

    """
    validate_dataframe(df)

    total_time_per_session = {}
    for session_id, session_df in df.groupby(ColumnNames.SESSION_ID):
        session_df = session_df.sort_values(by=ColumnNames.TIME_STAMP)
        total_time = session_df[ColumnNames.TIME_STAMP].iloc[-1] - session_df[ColumnNames.TIME_STAMP].iloc[0]
        total_time_per_session[session_id] = total_time
    return total_time_per_session

def movement_time(df: pd.DataFrame, traces: dict[str, list[pd.DataFrame]] = None) -> dict:
    """
    Calculate the total movement time from traces in milliseconds, taking pauses into account.
    This is the same as the interval of time the user is interacting with the interface.
    
    Parameters:
        df (pd.DataFrame): DataFrame containing 'timeStamp' column.
        traces (dict): Dictionary with sessionId as keys and list of DataFrames as values.
    Returns:
        dict: A dictionary with sessionId as keys and total movement time in milliseconds as values.
    """
    if traces is None:
        validate_dataframe(df)
        traces = extract_traces_by_session(df)

    movement_time_per_session = {}
    for session_id, session_traces in traces.items():
        total_movement_time = 0.0
        for trace in session_traces:
            trace = compute_space_time_diff(trace)  
            total_movement_time += trace[ColumnNames.DT].sum()
        movement_time_per_session[session_id] = total_movement_time

    return movement_time_per_session

def num_pauses(df: pd.DataFrame, threshold: float = 100, computeTraces: bool = True) -> tuple[dict, dict]:
    """
    Calculate the number of pauses in the DataFrame.
    
    Parameters:
        df (pd.DataFrame): DataFrame containing 'timeStamp' column.
        threshold (float): Time threshold in milliseconds to consider a pause, by default 100 ms.
        computeTraces (bool): Whether to compute traces by sessionId, by default True. If False, df is assumed to be already segmented by sessionId.
    
    Returns:
        tuple (tuple[dict, dict]): A tuple containing two dictionaries with the number of pauses per session and the mean number of pauses per trace, with the sessionId as keys.
    """

    validate_dataframe(df)
    
    if computeTraces:
        df = extract_traces_by_session(df)
    else:
        df_pauses = _num_pauses_trace(df, threshold)
        total_pauses_session = df_pauses.shape[0]
        metrics = {}
        metrics[df[ColumnNames.SESSION_ID].iloc[0]] = {
            "num_pauses": total_pauses_session,
            "mean_pauses_per_trace": total_pauses_session
        }
        return metrics

    metrics_per_session = {}
    for session_id, session_traces in df.items():
        total_pauses_session = 0
        for trace in session_traces:
            df_pauses = _num_pauses_trace(trace, threshold)
            total_pauses_session += df_pauses.shape[0]
        metrics_per_session[session_id] = {
            "num_pauses": total_pauses_session,
            "mean_pauses_per_trace": total_pauses_session / len(session_traces) if len(session_traces) > 0 else 0
        }
    return  metrics_per_session

def _num_pauses_trace(df: pd.DataFrame, threshold: float) -> pd.DataFrame:
    """
    Helper function to calculate pauses in a single trace.
    """
    df = df.sort_values(by=ColumnNames.TIME_STAMP).reset_index(drop=True)
    df[ColumnNames.DT] = df[ColumnNames.TIME_STAMP].diff().fillna(0)
    pauses = df[df[ColumnNames.DT] > threshold]
    return pauses


def pauses_metrics(df: pd.DataFrame, threshold: float = 100, traces: dict[str, list[pd.DataFrame]] = None) -> dict:
    """
    Calculate pause metrics for the given DataFrame.
    
    Parameters:
        df (pd.DataFrame): DataFrame containing 'timeStamp' column.
        threshold (float): Time threshold in milliseconds to consider a pause, by default 100 ms.
        traces (dict): Dictionary with sessionId as keys and list of DataFrames as values.

    Returns:
        dict: A dictionary with sessionId as keys and a dictionary of pause metrics as values.
    """

    if traces is None:
        validate_dataframe(df)  
        traces = extract_traces_by_session(df)
    
    pause_metrics_per_session = {}
    number_pauses_session, number_pauses_trace = num_pauses(traces, threshold, computeTraces=False)
    for session_id, session_traces in traces.items():
        total_pauses = 0
        total_pause_duration = 0.0
        pause_durations = []

        for trace in session_traces:
            trace = compute_space_time_diff(trace)  
            pauses = trace[trace[ColumnNames.DT] > threshold]
            total_pauses += pauses.shape[0]
            total_pause_duration += pauses[ColumnNames.DT].sum()
            pause_durations.extend(pauses[ColumnNames.DT].tolist())

        # Compute pause metrics for the session
        if total_pauses > 0:
            mean_pause_duration = total_pause_duration / total_pauses
        else:
            mean_pause_duration = 0

        pause_metrics_per_session[session_id] = {
            "total_pauses": total_pauses,
            "mean_pause_duration": mean_pause_duration,
            "pause_durations": pause_durations,
            "mean_pauses_per_trace": number_pauses_trace.get(session_id, 0),
            "max_pause": max(pause_durations) if pause_durations else 0,
            "min_pause": min(pause_durations) if pause_durations else 0,
        }

    return pause_metrics_per_session


"""
Utility functions for PyWib
"""
from .timing import execution_time, movement_time, num_pauses, pauses_metrics
from .movement import (velocity, acceleration, jerkiness, 
                       path, auc_optimal, auc_ratio, auc, auc_ratio_metrics,
                       velocity_metrics, acceleration_metrics, jerkiness_metrics,
                       deviations)
from .mouse import click_slip, number_of_clicks
from .keyboard import (typing_speed, typing_speed_metrics, backspace_usage)

__all__ = [
    "execution_time",
    "movement_time",
    "num_pauses",
    "pauses_metrics",
    "velocity",
    "acceleration",
    "jerkiness",
    "path",
    "auc_optimal",
    "auc_ratio",
    "auc_ratio_metrics",
    "auc",
    "velocity_metrics",
    "acceleration_metrics",
    "jerkiness_metrics",
    "click_slip",
    "number_of_clicks",
    "deviations"
    "typing_speed",
    "typing_speed_metrics",
    "backspace_usage",
]


import unittest
import numpy as np
import sys
import os
sys.path.insert(0, os.path.dirname(__file__))
from utils import process_csv, import_pyModule

import_pyModule()
from pywib import (velocity, acceleration, compute_space_time_diff, 
                   auc_ratio_metrics, auc_ratio, velocity_metrics, acceleration_metrics,
                   jerkiness, jerkiness_metrics)

# Cambiar a True solo al probar en desarrollo
DEBUG = True

class TestMovement(unittest.TestCase):
    
    def setUp(self):
        """Set up test data"""
        # Create sample test data instead of relying on external CSV
        if(DEBUG):
            self.test_data = process_csv('test/test_data/test_window_resize_error.csv')
            self.test_data_auc = process_csv('test/test_data/test_auc.csv')
        else:
            self.test_pause = process_csv('pywib/test/test_data/pauses.csv')
            self.test_data = process_csv('pywib/test/test_data/test_window_resize_error.csv')
            self.test_data_auc = process_csv('pywib/test/test_data/test_auc.csv')
        
    def test_velocity(self):
        """Test velocity calculation"""
        # Compute space-time differences
        df = compute_space_time_diff(self.test_data.copy())
        
        # Calculate velocity
        df_velocity = velocity(df, per_traces=True)
        
        # Check if the velocity column is added
        for _, traces in df_velocity.items():
            for trace in traces:
                self.assertIn('velocity', trace.columns)
                # Check if velocity values are non-negative
                self.assertTrue((trace['velocity'] >= 0).all())
                self.assertGreaterEqual(trace['velocity'].mean(), 0)
                # Verify velocity calculation
                dt = trace.timeStamp.diff()
                dx = trace.x.diff()
                dy = trace.y.diff()
                expected_velocity = np.sqrt(dx**2 + dy**2) / dt
                expected_velocity.fillna(0, inplace=True)
                np.testing.assert_allclose(trace['velocity'], expected_velocity, rtol=1e-5, atol=1e-8)

    def test_velocity_metrics(self):
        """Test velocity metrics calculation"""
        metrics = velocity_metrics(self.test_data.copy())

        for _, session_id in metrics.items():
            self.assertIn('mean', session_id)
            self.assertIn('max', session_id)
            self.assertIn('min', session_id)
            self.assertGreaterEqual(session_id['mean'], 0)
            self.assertGreaterEqual(session_id['max'], 0)
            self.assertGreaterEqual(session_id['min'], 0)

    def test_acceleration(self):
        """Test acceleration calculation"""
        df = compute_space_time_diff(self.test_data.copy())
        
        # Calculate acceleration
        df_acceleration = acceleration(df, per_traces=True)
        
        # Check if the acceleration column is added
        for _, traces in df_acceleration.items():
            for trace in traces:
                self.assertIn('acceleration', trace.columns)
                # Verify acceleration calculation
                dt = trace.timeStamp.diff()
                dv = trace['velocity'].diff()
                expected_acceleration = dv / dt
                expected_acceleration.fillna(0, inplace=True)
                np.testing.assert_allclose(trace['acceleration'], expected_acceleration, rtol=1e-5, atol=1e-8)

    def test_acceleration_from_traces(self):
        df = compute_space_time_diff(self.test_data.copy())
        acceleration_traces = acceleration(None, velocity(df, per_traces=True), per_traces=True)
        for _, traces in acceleration_traces.items():
            for trace in traces:
                self.assertIn('acceleration', trace.columns)
                # Verify acceleration calculation
                dt = trace.timeStamp.diff()
                dv = trace['velocity'].diff()
                expected_acceleration = dv / dt
                expected_acceleration.fillna(0, inplace=True)
                np.testing.assert_allclose(trace['acceleration'], expected_acceleration, rtol=1e-5, atol=1e-8)

    def test_acceleration_metrics(self):
        """Test acceleration metrics calculation"""
        acc_metrics = acceleration_metrics(self.test_data.copy())

        for _, session in acc_metrics.items():
            self.assertIn('mean', session)
            self.assertIn('max', session)
            self.assertIn('min', session)
            self.assertGreaterEqual(session['mean'], 0)
            self.assertGreaterEqual(session['max'], session['min'])

    def test_jerkiness(self):
        """Test jerkiness calculation"""
        jk_df = jerkiness(self.test_data.copy(), per_traces=True)

        for _, traces in jk_df.items():
            for trace in traces:
                self.assertIn('jerkiness', trace.columns)
                # Verify jerkiness calculation
                dt = trace.timeStamp.diff()
                da = trace['acceleration'].diff()
                expected_jerkiness = da / dt
                expected_jerkiness.fillna(0, inplace=True)
                np.testing.assert_allclose(trace['jerkiness'], expected_jerkiness, rtol=1e-5, atol=1e-8)

    def test_jerkiness_metrics(self):
        metrics = jerkiness_metrics(self.test_data.copy())

        for _, session in metrics.items():
            self.assertIn('mean', session)
            self.assertIn('max', session)
            self.assertIn('min', session)
            self.assertGreaterEqual(session['mean'], 0)
            self.assertGreaterEqual(session['max'], session['min'])

    def test_auc(self):
        auc = auc_ratio(self.test_data_auc.copy())
        for _, session in auc.items():
            for trace in session:
                self.assertIn('auc_ratio', trace)
                self.assertIn('auc', trace)
                self.assertGreaterEqual(trace['auc_ratio'], 0)
        
    def test_auc_metrics(self):
        auc_metrics = auc_ratio_metrics(self.test_data_auc.copy())
        for _, session in auc_metrics.items():
            self.assertIn('mean_ratio', session)
            self.assertIn('min_ratio', session)
            self.assertIn('max_ratio', session)
            self.assertGreaterEqual(session['mean_ratio'], 0)

if __name__ == '__main__':
    unittest.main()


import unittest
import numpy as np
import sys
import os
sys.path.insert(0, os.path.dirname(__file__))
from utils import process_csv, import_pyModule

import_pyModule()

from pywib import movement_time, execution_time

DEBUG = True

class TestData:
    window_resize_error_file = 'test/test_data/test_window_resize_error.csv'
    window_resize_error_execution_time = 18100  # in milliseconds

class TestTiming(unittest.TestCase):

    def setUp(self):
        """Set up test data"""
        # Create sample test data instead of relying on external CSV
        
        if(DEBUG):
            self.test_data = process_csv(TestData.window_resize_error_file)
        else:
            self.test_data = process_csv('pywib/' + TestData.window_resize_error_file)
        
    def test_execution_time(self):
        """Test execution_time function"""
        exec_time = execution_time(self.test_data)
        self.assertIsInstance(exec_time, dict)
        for session_id, time in exec_time.items():
            self.assertIsInstance(session_id, str)
            self.assertGreaterEqual(time, 0)
            np.testing.assert_allclose(time, TestData.window_resize_error_execution_time, rtol=1e-5, atol=1e-8)

    def test_movement_time(self):
        """Test movement_time function"""
        move_time = movement_time(self.test_data)
        self.assertIsInstance(move_time, dict)
        for session_id, time in move_time.items():
            self.assertIsInstance(session_id, str)
            self.assertIsInstance(time, float)
            self.assertLessEqual(time, TestData.window_resize_error_execution_time)


if __name__ == '__main__':
    unittest.main()


import pandas as pd
from pywib.utils import validate_dataframe, extract_mouse_click_traces_by_session_with_intial_pause
from pywib.core import auc_ratio

def obtain_straight_patterns(df: pd.DataFrame = None, traces : list[str, list[pd.DataFrame]] = None, per_traces= False, threshold = 100) -> list[pd.DataFrame]:
    """
        First described in 'Investigating the Differences in Web Browsing Behaviour of Chinese and European Users Using Mouse Tracking' (Lee & Chen, 2007),
        the Straight Pattern can be described as a direct or straight movement in direction to a target, characterized by a pause before a direct
        movement towards a target without significant pauses in between the initial movement and the target acquisition.
        
        TODO : initial pause

        Parameters:
            df (pd.DataFrame): DataFrame containing 'sessionId', 'sceneId', 'eventType', 'timeStamp', 'x', and 'y' columns.
            traces (list): List of traces to analyze. Each trace is a pd.DataFrame.
            per_traces (bool): If True, process per traces. Default is False.
            threshold (float): Threshold (in px) for the AUC perpendicular distance to consider a movement as a straight pattern. Default is 100.
        Returns:
            list[pd.DataFrame]: A list of DataFrames, each corresponding to a straight pattern trace.
    """
    if(traces is None and per_traces):
        validate_dataframe(df)
        # This variable will always contain point and click traces
        traces = extract_mouse_click_traces_by_session_with_intial_pause(df)

    if(not per_traces):
        raise NotImplementedError("The 'per_traces' functionality is not yet implemented.")
    
    auc_values = auc_ratio(None, traces= traces, per_traces= True)

    straight_patterns_per_session = {}
    for session_id, trace in auc_values.items():
        straight_patterns = []
        for i, auc in enumerate(trace):
            # TODO usar auc_ratio con un porcentaje del threshold?
            if(auc['auc_perp'] <= threshold):
                straight_patterns.append(traces[session_id][i])
        straight_patterns_per_session[session_id] = straight_patterns
    return straight_patterns_per_session


import pandas as pd
import numpy as np
from ..utils.validation import validate_dataframe
from ..utils.utils import compute_space_time_diff
from ..utils.segmentation import extract_traces_by_session
from ..constants import ColumnNames, EventTypes

def number_of_clicks(df: pd.DataFrame) -> dict:
    """
    Calculate the number of clicks per session.
    Parameters:
        df (pd.DataFrame): DataFrame containing mouse event data.
    Returns:
        dict: A dictionary with session IDs as keys and number of clicks as values.
    """
    validate_dataframe(df)

    clicks_per_session = {}
    df = df.groupby(ColumnNames.SESSION_ID)
    for session_id, group in df:
        clicks_per_session[session_id] = group[group[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_CLICK].shape[0]
    return clicks_per_session

def click_slip(df: pd.DataFrame, threshold: float = 5.0) -> dict:
    """
    Calculate the number of click slips per session.
    A click slip is defined as a click event that occurs within a certain distance
    from the previous mouse position (indicating an unintended click).

    Parameters:
        df (pd.DataFrame): DataFrame containing mouse event data.
        threshold (float): Distance threshold to consider a click as a slip.

    Returns:
        dict: A dictionary with session IDs as keys and the metrics (click slips, max, min, average) as values.
    """
    validate_dataframe(df)

    click_slips_per_session = {}
    df = df.groupby(ColumnNames.SESSION_ID)
    metrics_per_session = {}
    for session_id, group in df:
        group = group.sort_values(by=ColumnNames.TIME_STAMP)
        slips = 0
        # We define a click slip as: the total path distance of all
        # EVENT_ON_MOUSE_MOVE events that occur between an
        # EVENT_ON_MOUSE_DOWN and the subsequent EVENT_ON_MOUSE_UP.
        # If that total move-distance is less than `threshold` we count
        # it as a slip.
        in_down = False
        last_move_x = None
        last_move_y = None
        accumulated_move_distance = 0.0
        distances = []
        durations = []
        mouse_down_time = None
        for _, row in group.iterrows():
            # Mouse down: start a new segment
            if row[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_MOUSE_DOWN:
                in_down = True
                # reset accumulators
                last_move_x = row[ColumnNames.X]
                last_move_y = row[ColumnNames.Y]
                accumulated_move_distance = 0.0
                mouse_down_time = row[ColumnNames.TIME_STAMP]

            # While between down and up, accumulate move distances
            if in_down and row[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_MOUSE_MOVE:
                x = row[ColumnNames.X]
                y = row[ColumnNames.Y]
                if last_move_x is not None and last_move_y is not None:
                    d = np.hypot(x - last_move_x, y - last_move_y)
                    accumulated_move_distance += d
                # set last seen move position (even if it was the first move)
                last_move_x, last_move_y = x, y

            # Mouse up: finalize the segment and decide if it's a slip
            if row[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_ON_MOUSE_UP and in_down:
                x = row[ColumnNames.X]
                y = row[ColumnNames.Y]
                duration = row[ColumnNames.TIME_STAMP] - mouse_down_time if mouse_down_time is not None else 0
                if last_move_x is not None and last_move_y is not None:
                    d = np.hypot(x - last_move_x, y - last_move_y)
                    accumulated_move_distance += d
                if accumulated_move_distance >= threshold:
                    slips += 1
                    distances.append(accumulated_move_distance)
                durations.append(duration)
                in_down = False
                last_move_x = None
                last_move_y = None
                mouse_down_time = None
                accumulated_move_distance = 0.0
        click_slips_per_session[session_id] = {
            "slips": slips,
            "distances": distances,
        }
        metrics = {
            "click_slips": slips,
            "longest_click_slip": max(distances) if distances else 0,
            "shortest_click_slip": min(distances) if distances else 0,
            "average_click_slip": slips / len(distances) if distances else 0,
            "average_click_slip_distance": np.mean(distances) if distances else 0,
            "average_click_duration": np.mean(durations) if durations else 0,
            "max_click_duration": max(durations) if durations else 0,
            "min_click_duration": min(durations) if durations else 0,
        }
        metrics_per_session[session_id] = metrics
    return metrics_per_session


import os
import sys
sys.path.insert(0, os.path.abspath('../../src'))
import pybtex.plugin
from pybtex.style.formatting.unsrt import Style as UnsrtStyle
from pybtex.style.template import (
    names, sentence, field, optional, words, 
)

# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = 'pywib'
copyright = '2025, Guillermo Dylan Carvajal Aza, Alejandro Álvarez Varela'
author = 'Guillermo Dylan Carvajal Aza, Alejandro Álvarez Varela'
release = '0.1.0'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = []

templates_path = ['_templates']
exclude_patterns = []

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.mathjax",
    "sphinx_autodoc_typehints",
    "sphinxcontrib.bibtex",
    "myst_parser",
]

# When True (default) Sphinx prefixes documented object names with the
# full module path. Set to False to show just the object name in
# autodoc/autosummary signatures 
add_module_names = False

# Bibliography for sphinxcontrib-bibtex
bibtex_bibfiles = ['references.bib']

class APAStyle(UnsrtStyle):
    def format_article(self, e):
        template = sentence [
            names('author', sep=', ', sep2=', ', last_sep=', '),  
            field('year', prefix='. (', suffix='). '),           
            field('title', suffix='. '),                          
            field('journal', suffix='. '),
            optional [ field('volume', prefix='Vol. ', suffix='. ') ],
            optional [ field('pages', prefix='pp. ', suffix='. ') ],
            optional [ field('url', prefix='', suffix='') ],
        ]
        return template.format_data(e)

    def format_incollection(self, e):
        template = sentence [
            names('author', sep=', ', sep2=', ', last_sep=', '),
            field('year', prefix='. (', suffix=').' ),
            field('title', suffix='. '),
            words [
                field('booktitle', prefix='In ', suffix=''),
                optional [ field('edition', prefix=' (', suffix=' ed.)') ],
                optional [ field('pages', prefix=', pp. ', suffix='') ],
                field('publisher', prefix='. ', suffix='.'),
            ],
            optional [ field('url', prefix=' ', suffix='') ],
        ]
        return template.format_data(e)

    def format_book(self, e):
        template = sentence [
            names('author', sep=', ', sep2=', ', last_sep=', '),
            field('year', prefix='. (', suffix='). '),
            field('title', suffix='. '),
            field('publisher', suffix='. '),
            optional [ field('edition', suffix=' ed. ') ],
            optional [ field('pages', prefix='pp. ', suffix='. ') ],
            optional [ field('url', prefix='', suffix='') ],
        ]
        return template.format_data(e)

    def format_labels(self, sorted_entries):
        for entry in sorted_entries:
            label = self.format_label(entry)
            yield [label]
    
pybtex.plugin.register_plugin('pybtex.style.formatting', 'apa', APAStyle)

# Use your new style for sphinxcontrib-bibtex
bibtex_default_style = 'apa'
bibtex_reference_style = 'author_year'

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'furo'
html_favicon = 'favicon.ico'
html_static_path = ['_static']

def setup(app):
    app.add_css_file("custom.css")


import unittest
import numpy as np
import sys
import os
sys.path.insert(0, os.path.dirname(__file__))
from utils import process_csv, import_pyModule

import_pyModule()

from pywib import number_of_clicks, click_slip

DEBUG = True

class TestData:
    
    dataFile = 'test/test_data/test_mouse.csv'
    expected_clicks = {
        'SESSION_A': 2,
        'SESSION_B': 2
    }
    expected_click_slips = {
        'SESSION_A': {'click_slips': 0,'longest_click_slip':0, 'shortest_click_slip': 0, 'average_click_slip': 0, 'average_click_slip_distance': 0, 'average_click_duration':  np.float64(30.0), 'max_click_duration':  np.float64(30.0), 'min_click_duration':  np.float64(30.0)},
        'SESSION_B': {'click_slips': 1, 'longest_click_slip': np.float64(25.0), 'shortest_click_slip': np.float64(25.0), 'average_click_slip': 1.0, 'average_click_slip_distance': np.float64(25.0), 'average_click_duration': np.float64(30.0), 'max_click_duration': np.float64(30.0), 'min_click_duration': np.float64(30.0)}
    }

class TestMouse(unittest.TestCase):
    
    def setUp(self):
        """Set up test data"""
        # Create sample test data instead of relying on external CSV
        if(DEBUG):
            self.test_data = process_csv(TestData.dataFile)
        else:
            self.test_data = process_csv('pywib/' + TestData.dataFile)
        
    def test_number_of_clicks(self):
        clicks = number_of_clicks(self.test_data)
        self.assertEqual(clicks, TestData.expected_clicks)

    def test_click_slip(self):
        click_slips = click_slip(self.test_data)
        self.assertEqual(click_slips['SESSION_A'], TestData.expected_click_slips['SESSION_A'])
        self.assertEqual(click_slips['SESSION_B'], TestData.expected_click_slips['SESSION_B'])

    def test_click_duration(self):
        durations = click_slip(self.test_data)
        self.assertIn('SESSION_A', durations)
        self.assertIn('SESSION_B', durations)
        self.assertGreaterEqual(durations['SESSION_A']['average_click_duration'], 30)
        self.assertGreaterEqual(durations['SESSION_B']['average_click_duration'], 30)
        self.assertGreaterEqual(durations['SESSION_A']['max_click_duration'], durations['SESSION_B']['min_click_duration'])
        self.assertGreaterEqual(durations['SESSION_B']['max_click_duration'], durations['SESSION_B']['min_click_duration'])


if __name__ == '__main__':
    unittest.main()


import pandas as pd
from ..constants import ColumnNames

required_columns = [
    ColumnNames.SESSION_ID, ColumnNames.EVENT_TYPE, ColumnNames.TIME_STAMP, ColumnNames.X, ColumnNames.Y
]

keyboard_columns = [
   ColumnNames.KEY_VALUE_EVENT, ColumnNames.KEY_CODE_EVENT
]

def validate_dataframe(df: pd.DataFrame):

    for col in required_columns:
        if col not in df.columns:
            raise ValueError(f"Missing required column: {col}")
        
def validate_dataframe_keyboard(df: pd.DataFrame):

    validate_dataframe(df)

    columns_to_check = keyboard_columns

    for col in columns_to_check:
        if col not in df.columns:
            raise ValueError(f"Missing required column: {col}")
        
def validate_duplicate_timestamps(df: pd.DataFrame):
    for session_id, group in df.groupby(ColumnNames.SESSION_ID):
        if group[ColumnNames.TIME_STAMP].duplicated().any():
            raise ValueError(f"Duplicate timestamps found in session {session_id}")


"""
Utility functions for PyWib
"""

from .validation import validate_dataframe, validate_dataframe_keyboard, validate_duplicate_timestamps
from .segmentation import extract_traces_by_session, extract_mouse_click_traces_by_ession, extract_mouse_click_traces_by_session_with_intial_pause
from .visualization import visualize_trace, video_from_trace
from .utils import compute_space_time_diff, compute_metrics_from_traces
from .movement import (acceleration_traces, velocity_traces, velocity_df, 
                       acceleration_df, jerkiness_df, jerkiness_traces, _path,
                       auc_ratio_traces, auc_ratio_df)

__all__ = [
    'validate_dataframe',
    'validate_dataframe_keyboard',
    'extract_traces_by_session',
    'visualize_trace',
    'compute_space_time_diff',
    'acceleration_traces',
    'jerkiness_traces',
    'velocity_traces',
    'velocity_df',
    'acceleration_df',
    'jerkiness_df',
    '_path',
    'compute_metrics_from_traces',
    'auc_ratio_traces',
    'auc_ratio_df',
    'extract_mouse_click_traces_by_ession',
    'extract_mouse_click_traces_by_session_with_intial_pause',
    'video_from_trace',
    'validate_duplicate_timestamps'
]



import pandas as pd
import numpy as np
from pywib.constants import ColumnNames, EventTypes, KeyCodeEvents
from pywib.utils.validation import validate_dataframe_keyboard 

def typing_durations_df(df: pd.DataFrame = None, validate: bool = True) -> list:
    """
    Calculate the durations of individual keystrokes from a DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing interaction data with 'event_type', 'timestamp and 'key' columns.
        validate (bool): Whether to validate the input DataFrame. Default is True.
    Returns:
        list: List of keystroke durations in milliseconds.
    """
    if validate:
        validate_dataframe_keyboard(df)
    durations = []

    key_down_events = df[df[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_KEY_DOWN]
    key_up_events = df[df[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_KEY_UP]
    for key in key_down_events[ColumnNames.KEY].unique():
        down_times = key_down_events[key_down_events[ColumnNames.KEY] == key][ColumnNames.TIME_STAMP].values
        up_times = key_up_events[key_up_events[ColumnNames.KEY] == key][ColumnNames.TIME_STAMP].values
        paired_times = zip(down_times, up_times)
        for down_time, up_time in paired_times:
            duration = up_time - down_time
            if duration >= 0:
                durations.append(duration)
    return durations

def typing_durations_traces(traces: dict[str, list[pd.DataFrame]], validate: bool = True) -> dict[str, list[pd.DataFrame]]:
    """
    Calculate the durations of individual keystrokes from pre-extracted keystroke traces.

    Parameters:
        traces (dict[str, list[pd.DataFrame]]): Pre-extracted keystroke traces by session.
        validate (bool): Whether to validate the input DataFrames. Default is True.
    Returns:
        dict (dict[str, list[pd.DataFrame]]): A dictionary with session IDs as keys and lists of keystroke durations per trace as values.
    """
    durations_per_session = {}
    for session_id, keystroke_traces in traces.items():
        durations = []
        for trace in keystroke_traces:
            trace_durations = typing_durations_df(trace, validate)
            durations.append(trace_durations)
        durations_per_session[session_id] = durations
    return durations_per_session

def typing_speed_df(df: pd.DataFrame = None, validate: bool = True) -> float:
    """
    Calculate the average typing speed in characters per minute (CPM) from a DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing interaction data with 'event_type', 'timestamp', and 'key' columns.
        validate (bool): Whether to validate the input DataFrame. Default is True.
    Returns:
        float: Average typing speed in characters per minute (CPM).
    """
    
    if validate:
        validate_dataframe_keyboard(df)
    total_chars = 0
    total_time = 0.0  # in seconds
    strokes = df[df[ColumnNames.EVENT_TYPE].isin([EventTypes.EVENT_KEY_UP, EventTypes.EVENT_KEY_DOWN])]
    keys = df[df[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_KEY_UP]
    total_chars += len(keys)
    if not strokes.empty:
        time_start = strokes[ColumnNames.TIME_STAMP].iloc[0]
        time_end = strokes[ColumnNames.TIME_STAMP].iloc[-1]
        total_time = (time_end - time_start) / 1000.0  # convert ms to s
        if total_time > 0:
            cpm = (total_chars / total_time) * 60.0
            return cpm
    return 0.0

def typing_speed_traces(traces: dict[str, list[pd.DataFrame]], validate: bool = True) -> dict[str, list[float]]:
    """
    Calculate the average typing speed in characters per minute (CPM) from pre-extracted keystroke traces.

    Parameters:
        traces (dict[str, list[pd.DataFrame]]): Pre-extracted keystroke traces by session.
        validate (bool): Whether to validate the input DataFrames. Default is True.
    Returns:
        dict (dict[str, list[float]]): A dictionary with session IDs as keys and lists of typing speeds (CPM) per trace as values.
    """
    speed_by_session = {}
    for session_id, keystroke_traces in traces.items():
        speed_by_trace = []
        for trace in keystroke_traces:
            cpm = typing_speed_df(trace, validate)
            speed_by_trace.append(cpm)
        speed_by_session[session_id] = speed_by_trace
    return speed_by_session

def backspace_usage_df(df: pd.DataFrame = None, validate: bool = True) -> float:
    """
    Calculate the backspace usage rate as a percentage of total keystrokes from a DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing interaction data with 'event_type', 'timestamp', and 'key' columns.
        validate (bool): Whether to validate the input DataFrame. Default is True.
    Returns:
        float: Backspace usage rate as a percentage of total keystrokes.
    """
    if validate:
        validate_dataframe_keyboard(df)
    mask = (df[ColumnNames.EVENT_TYPE] == EventTypes.EVENT_KEY_DOWN) & ( (df[ColumnNames.KEY_CODE_EVENT] == KeyCodeEvents.KEY_CODE_BACKSPACE) | (df[ColumnNames.KEY_CODE_EVENT] == KeyCodeEvents.KEY_CODE_DELETE))

    return int(mask.sum())

def backspace_usage_traces(traces: dict[str, list[pd.DataFrame]], validate: bool = True) -> dict[str, list[float]]:
    """
    Calculate the backspace usage rate as a percentage of total keystrokes from pre-extracted keystroke traces.

    Parameters:
        traces (dict[str, list[pd.DataFrame]]): Pre-extracted keystroke traces by session.
        validate (bool): Whether to validate the input DataFrames. Default is True.
    Returns:
        dict (dict[str, list[float]]): A dictionary with session IDs as keys and lists of backspace usage rates per trace as values.
    """
    usage_by_session = {}
    for session_id, keystroke_traces in traces.items():
        backspace_count = 0
        for trace in keystroke_traces:
            usage = backspace_usage_df(trace, validate)
            backspace_count += usage
        usage_by_session[session_id] = backspace_count
    return usage_by_session


import unittest
import numpy as np
import sys
import os
import pytest
sys.path.insert(0, os.path.dirname(__file__))
from utils import process_csv, import_pyModule, csv_to_df_no_checks
import_pyModule()

from pywib import validate_duplicate_timestamps, validate_dataframe, validate_dataframe_keyboard

DEBUG = True

class TestData:
    if(DEBUG):
        duplicateFile = 'test/test_data/test_duplicate_timestamps.csv'
        incorrectColumns = 'test/test_data/test_incorrect_columns.csv'
        keyboardMissing = 'test/test_data/test_missing_keyboard_cols.csv'
    else:
        duplicateFile = 'pywib/test/test_data/test_duplicate_timestamps.csv'
        incorrectColumns = 'pywib/test/test_data/test_incorrect_columns.csv'
        keyboardMissing = 'pywib/test/test_data/test_missing_keyboard_cols.csv'

class TestValidation(unittest.TestCase):

    def test_duplicate_timestamps_should_fail(self):
        duplicate_data = process_csv(TestData.duplicateFile)
        try:
            validate_duplicate_timestamps(duplicate_data)
            self.fail("validate_no_duplicate_timestamps did not raise an exception for duplicate timestamps")
        except Exception as e:
            self.assertEqual(str(e.args[0]), "Duplicate timestamps found in session SESSION_A")

    def test_validate_dataframe_should_fail(self):
        incorrect_data = csv_to_df_no_checks(TestData.incorrectColumns)
        try:
            validate_dataframe(incorrect_data)
            self.fail("validate_dataframe did not raise an exception for duplicate timestamps")
        except Exception as e:
            self.assertEqual(str(e.args[0]), "Missing required column: sessionId")

    def test_validate_dataframe_keyboard_should_fail(self):
        incorrect_data = csv_to_df_no_checks(TestData.keyboardMissing)
        try:
            validate_dataframe_keyboard(incorrect_data)
            self.fail("validate_dataframe_keyboard did not raise an exception for duplicate timestamps")
        except Exception as e:
            self.assertEqual(str(e.args[0]), "Missing required column: keyValueEvent")


from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("requirements.txt", "r", encoding="utf-8") as fh:
    requirements = [line.strip() for line in fh if line.strip() and not line.startswith("#")]


setup(
    name='pywib',
    packages=find_packages(),
    version='0.1.0',
    description='HCI Web Interaction Analyzer - A library for analyzing web user interactions',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/HumanCommunicationInteraction/pywib",
    packages=find_packages(),
    author='Guillermo Dylan Carvajal Aza',
    author_email='carvajalguillermo@uniovi.es',
    keywords=['HCI', 'Web Interaction', 'Analyzer'],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Topic :: Scientific/Engineering",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    python_requires=">=3.9",
    install_requires=requirements,
    extras_require={
        "dev": [
            "pytest>=6.0",
            "pytest-cov",
            "black",
            "flake8",
        ],
    },
    include_package_data=True,
    zip_safe=False,
)


import pandas as pd
import numpy as np
from pywib.constants import ColumnNames
from pywib.utils import compute_space_time_diff, validate_dataframe

def velocity_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculate velocity for a single DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'x', 'y', and 'timeStamp' columns.

    Returns:
        pd.DataFrame: DataFrame with an additional 'velocity' column.
    """
    validate_dataframe(df)
    df = _path(df)  # Compute distance and dt first
    df['velocity'] = df['distance'] / df['dt']
    df['velocity'] = df['velocity'].fillna(0)  # Replace NaN (first point) with 0
    return df


def velocity_traces(traces: dict[str, list[pd.DataFrame]]) -> dict[str, list[pd.DataFrame]]:
    """
    Calculate velocity for a dictionary of traces (each a list of DataFrames).

    Parameters:
        traces (dict[str, list[pd.DataFrame]]): Mapping of sessionId to list of DataFrames.

    Returns:
        dict[str, list[pd.DataFrame]]: Same structure, but with velocity computed in each DataFrame.
    """
    for session_id, session_traces in traces.items():
        for i, df in enumerate(session_traces):
            validate_dataframe(df)
            session_traces[i] = velocity_df(df)
        traces[session_id] = session_traces
    return traces

import pandas as pd

def acceleration_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculate acceleration for a single DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'x', 'y', and 'timeStamp' columns.

    Returns:
        pd.DataFrame: DataFrame with an additional 'acceleration' column.
    """
    validate_dataframe(df)

    if(ColumnNames.VELOCITY not in df.columns):
        df = velocity_df(df)

    df['acceleration'] = df['velocity'].diff().fillna(0) / df['dt']
    df['acceleration'] = df['acceleration'].fillna(0)
    return df


def acceleration_traces(traces: dict[str, list[pd.DataFrame]]) -> dict[str, list[pd.DataFrame]]:
    """
    Calculate acceleration for a dictionary of traces (each a list of DataFrames).

    Parameters:
        traces (dict[str, list[pd.DataFrame]]): Mapping of sessionId to list of DataFrames.

    Returns:
        dict[str, list[pd.DataFrame]]: Same structure, but with acceleration computed in each DataFrame.
    """
    for session_id, session_traces in traces.items():
        for i, df in enumerate(session_traces):
            validate_dataframe(df)
            session_traces[i] = acceleration_df(df)
        traces[session_id] = session_traces
    return traces

def jerkiness_df(df: pd.DataFrame) -> pd.DataFrame:
    """
    Calculate jerkiness for a single DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'x', 'y', and 'timeStamp' columns.

    Returns:
        pd.DataFrame: DataFrame with an additional 'jerkiness' column.
    """
    validate_dataframe(df)

    if(ColumnNames.ACCELERATION not in df.columns):
        df = acceleration_df(df)

    df['jerkiness'] = df['acceleration'].diff().fillna(0) / df['dt']
    df['jerkiness'] = df['jerkiness'].fillna(0)
    return df


def jerkiness_traces(traces: dict[str, list[pd.DataFrame]]) -> dict[str, list[pd.DataFrame]]:
    """
    Calculate jerkiness for a dictionary of traces (each a list of DataFrames).

    Parameters:
        traces (dict[str, list[pd.DataFrame]]): Mapping of sessionId to list of DataFrames.

    Returns:
        dict[str, list[pd.DataFrame]]: Same structure, but with jerkiness computed in each DataFrame.
    """
    for session_id, session_traces in traces.items():
        for i, df in enumerate(session_traces):
            validate_dataframe(df)
            session_traces[i] = jerkiness_df(df)
        traces[session_id] = session_traces
    return traces


def jerkiness(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None) -> dict[str, list[pd.DataFrame]]:
    """
    Compute jerkiness for either a single DataFrame or multiple traces.

    If `traces` is not provided, they are extracted from `df` using `extract_traces_by_session()`.

    Parameters
    ----------
    df : pd.DataFrame, optional
        DataFrame containing 'acceleration' and 'dt' columns.
    traces : dict[str, list[pd.DataFrame]], optional
        Dictionary mapping session IDs to lists of DataFrames.

    Returns
    -------
    dict[str, list[pd.DataFrame]]
        Dictionary of traces, each containing the computed 'jerkiness' column.
    """
    if traces is None:
        if df is None:
            raise ValueError("Either 'df' or 'traces' must be provided.")
        validate_dataframe(df)
        traces = extract_traces_by_session(df)
    return jerkiness_traces(traces)

def _path(trace: pd.DataFrame) -> pd.DataFrame:
    """
    Helper function to calculate the path length for a single trace.
    This function computes the path length based on the Euclidean distance between consecutive points.

    Parameters:
        trace (pd.DataFrame): A single trace DataFrame.

    Returns:
        pd.DataFrame: DataFrame with an additional 'distance' column representing the path length.
    """

    if trace is None:
        raise ValueError("Trace DataFrame must be provided.")

    validate_dataframe(trace)

    trace = compute_space_time_diff(trace)
    trace['distance'] = np.sqrt(trace['dx'] ** 2 + trace['dy'] ** 2)

    return trace

def _auc(df: pd.DataFrame) -> float:
    """
    Helper function to calculate the Area Under the Curve (AUC) for a single DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'x' and 'y' columns.
    Returns:
        float: The computed AUC value.
    """
    validate_dataframe(df)

    df = compute_space_time_diff(df)

    # Área bajo la curva real
    area_real = np.trapezoid(df[ColumnNames.Y], df[ColumnNames.X])
    return area_real

def _auc_optimal(df: pd.DataFrame) -> float:
    """
    Helper function to calculate the Area Under the Optimal Curve for a single DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'x' and 'y' columns.
    Returns:
        float: The computed Area Under the Optimal Curve value.
    """
    df = compute_space_time_diff(df)

    # Área bajo la línea óptima
    x0, y0 = df[ColumnNames.X].iloc[0], df[ColumnNames.Y].iloc[0]
    x1, y1 = df[ColumnNames.X].iloc[-1], df[ColumnNames.Y].iloc[-1]
    x_opt = np.linspace(x0, x1, len(df))
    y_opt = np.linspace(y0, y1, len(df))
    area_optimal = np.trapezoid(y_opt, x_opt)

    return area_optimal

def auc_ratio_df(df: pd.DataFrame) -> dict:
    """
    Calculate AUC and AUC ratio for a single DataFrame, returning them as a dictionary.
    
    Parameters:
        df (pd.DataFrame): DataFrame containing 'x' and 'y' columns.
    Returns:
        dict: Dictionary with 'auc', and 'auc_ratio'.
    """
    validate_dataframe(df)
    auc_val = _auc(df)
    area_optimal = _auc_optimal(df)
    df_opt = compute_optimal_path(df)
    # Prepare arrays
    user_x, user_y = df['x'].values, df['y'].values
    opt_x, opt_y = df_opt['x'].values, df_opt['y'].values

    # Compute perpendicular distances from optimal points to user segments
    dists = []
    for i in range(len(opt_x)):
        px, py = opt_x[i], opt_y[i]
        # Compute distance to all user segments and take minimum
        seg_dists = [
            point_to_segment_distance(px, py, user_x[j], user_y[j], user_x[j+1], user_y[j+1])
            for j in range(len(user_x)-1)
        ]
        dists.append(min(seg_dists))

    # Integrate along optimal path
    # Compute optimal path length increments (arc-length)
    dx = np.diff(opt_x)
    dy = np.diff(opt_y)
    ds = np.hypot(dx, dy)
    s_grid = np.concatenate(([0], np.cumsum(ds)))

    auc_perp = np.trapezoid(dists, s_grid)
    
    # Optional normalization by total optimal path length TODO correct?
    total_opt_length = s_grid[-1]
    if total_opt_length > 0:
        auc_perp /= total_opt_length

    return {
        "auc": auc_val,
        "auc_ratio": abs(auc_val - area_optimal) / (abs(area_optimal) + 1e-6),
        'auc_perp': auc_perp
    }

def compute_optimal_path(df: pd.DataFrame, n_points: int = 100) -> pd.DataFrame:
    """
    Compute an "optimal" trajectory as a straight line between start and end points
    of the user path, sampled uniformly with n_points.

    Parameters:
        df (pd.DataFrame): User trajectory with 'x' and 'y' columns.
        n_points (int): Number of points in the optimal path.

    Returns:
        pd.DataFrame: DataFrame with columns 'x' and 'y' representing the optimal path.
    """
    start_x, start_y = df['x'].iloc[0], df['y'].iloc[0]
    end_x, end_y = df['x'].iloc[-1], df['y'].iloc[-1]

    x_opt = np.linspace(start_x, end_x, n_points)
    y_opt = np.linspace(start_y, end_y, n_points)

    df_opt = pd.DataFrame({'x': x_opt, 'y': y_opt})
    return df_opt

def point_to_segment_distance(px, py, x1, y1, x2, y2):
        """Compute the shortest distance from point (px,py) to segment [(x1,y1),(x2,y2)]"""
        # Vector from x1,y1 to px,py
        dx, dy = x2 - x1, y2 - y1
        if dx == dy == 0:
            # Segment is a point
            return np.hypot(px - x1, py - y1)
        # Project point onto segment, computing parameter t
        t = ((px - x1) * dx + (py - y1) * dy) / (dx*dx + dy*dy)
        t = np.clip(t, 0, 1)
        closest_x = x1 + t * dx
        closest_y = y1 + t * dy
        return np.hypot(px - closest_x, py - closest_y)

def auc_ratio_traces(traces: dict[str, list[pd.DataFrame]]) -> dict[list[dict]]:
    """
    Calculate AUC ratio metrics for multiple traces grouped by session IDs.

    Parameters:
        traces (dict[str, list[pd.DataFrame]]): Mapping of sessionId to list of DataFrames.
    Returns:
        dict[list[dict]]: Mapping of sessionId to list of AUC ratio metrics dictionaries.
    """
    auc_metrics = {}
    for session_id, session_traces in traces.items():
        auc_metrics_per_trace = []
        for i, df in enumerate(session_traces):
            validate_dataframe(df)
            auc_metrics_per_trace.append(auc_ratio_df(df))
        auc_metrics[session_id] = auc_metrics_per_trace
    return auc_metrics




import pandas as pd
import numpy as np

from pywib.utils import (acceleration_traces, velocity_traces, velocity_df, 
                         acceleration_df, jerkiness_df, jerkiness_traces, 
                         _path, validate_dataframe, compute_space_time_diff, 
                         compute_metrics_from_traces, extract_traces_by_session, 
                         auc_ratio_traces, auc_ratio_df)
from pywib.constants import ColumnNames

def velocity(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None, per_traces: bool = False) -> dict[str, list[pd.DataFrame]]:
    """
    Function to calculate velocity for either a single DataFrame or a traces dictionary.

    If `traces` is None, they will be computed from the DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'x', 'y', and 'timeStamp' columns.
        traces (dict[str, list[pd.DataFrame]]): Dictionary mapping session IDs to lists of DataFrames.
        per_traces (bool): Whether to compute velocity per trace. If False, compute directly on df.

    Returns:
        dict[str, list[pd.DataFrame]]: Dictionary of traces with computed 'velocity' column.
    """
    if df is None and traces is None:
        raise ValueError("Either 'df' or 'traces' must be provided.")

    if not per_traces:
        # Compute directly on the DataFrame (no trace extraction)
        return velocity_df(df)

    # If traces are not provided, extract them from df
    if traces is None:
        validate_dataframe(df)
        traces = extract_traces_by_session(df)

    # Compute velocity for each trace
    return velocity_traces(traces)


def velocity_metrics(df: pd.DataFrame, traces: dict[str, list[pd.DataFrame]] = None) -> dict:
    """
    Calculate velocity metrics for the given DataFrame or traces.
    This function computes the mean, max, and min velocity for each session.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'velocity' column.
        traces (dict): A dictionary with keys as (sessionId) and values as lists of DataFrames. If None, traces will be computed from df.

    Returns:
        dict: A dictionary with keys as (sessionId) and values as dictionaries with 'mean ', 'max', and 'min' velocity.
    """
    
    return compute_metrics_from_traces(
        df=df,
        traces=traces,
        column_name=ColumnNames.VELOCITY,
        compute_traces_fn=velocity,
        preprocess_fn=lambda s: s[s > 0]  # Exclude zero velocities
    )

def acceleration(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None, per_traces: bool = False) -> dict[str, list[pd.DataFrame]]:
    """
    Wrapper function to calculate acceleration for either a single DataFrame or a traces dictionary.

    If `traces` is None, they will be computed from the DataFrame.
    """
    if df is None and traces is None:
        raise ValueError("Either 'df' or 'traces' must be provided.")

    if not per_traces:
        # Compute directly on the DataFrame (no trace extraction)
        return acceleration_df(df)

    # If traces are not provided, extract them from df
    if traces is None:
        validate_dataframe(df)
        traces = extract_traces_by_session(df)

    # Compute acceleration for each trace
    return acceleration_traces(traces)

def acceleration_metrics(df: pd.DataFrame, traces: dict[str, list[pd.DataFrame]] = None) -> dict:
    """
    Calculate acceleration metrics for the given DataFrame or traces.
    This function computes the mean, max, and min acceleration for each session.

    Parameters:
        df (pd.DataFrame): DataFrame containing interaction data. Optionally already including 'acceleration' column.
        traces (dict): A dictionary with keys as (sessionId) and values as lists of DataFrames. If None, traces will be computed from df.
    Returns:
        dict: A dictionary with keys as (sessionId) and values as dictionaries with 'mean', 'max', and 'min' acceleration.
    """
    if (ColumnNames.ACCELERATION not in df.columns) and (traces is None):
        validate_dataframe(df)
        if ColumnNames.VELOCITY not in df.columns:
            traces = velocity(df, per_traces=True)
        traces = acceleration(None, traces,  per_traces=True)

    return compute_metrics_from_traces(
        df=df,
        traces=traces,
        column_name=ColumnNames.ACCELERATION,
        compute_traces_fn=lambda _: traces,  # Already computed above
        preprocess_fn=lambda s: s[s != 0]    # Exclude zero accelerations
    )

def jerkiness(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None, per_traces: bool = False) -> dict[str, list[pd.DataFrame]]:
    """
    Compute jerkiness for either a single DataFrame or multiple traces.

    If `traces` is not provided, they are extracted from `df` using `extract_traces_by_session()`.

    Parameters
    ----------
    df : pd.DataFrame, optional
        DataFrame containing 'acceleration' and 'dt' columns.
    traces : dict[str, list[pd.DataFrame]], optional
        Dictionary mapping session IDs to lists of DataFrames.

    Returns
    -------
    dict[str, list[pd.DataFrame]]
        Dictionary of traces, each containing the computed 'jerkiness' column.
    """
    if df is None and traces is None:
        raise ValueError("Either 'df' or 'traces' must be provided.")

    if not per_traces:
        # Compute directly on the DataFrame (no trace extraction)
        return jerkiness_df(df)

    # If traces are not provided, extract them from df
    if traces is None:
        validate_dataframe(df)
        traces = extract_traces_by_session(df)

    # Compute jerkiness for each trace
    return jerkiness_traces(traces)

def jerkiness_metrics(df: pd.DataFrame, traces: dict[str, list[pd.DataFrame]] = None) -> dict:
    """
    Calculate jerkiness metrics for the given DataFrame or traces.
    This function computes the mean, max, and min jerkiness for each session.

    Parameters:
        df (pd.DataFrame): DataFrame containing interaction data. Optionally already including 'jerkiness' column.
        traces (dict): A dictionary with keys as (sessionId) and values as lists of DataFrames. If None, traces will be computed from df.
    Returns:
        dict: A dictionary with keys as (sessionId) and values as dictionaries with 'mean', 'max', and 'min' jerkiness.
    """
    if((ColumnNames.JERKINESS not in df.columns) and (traces is None)):
        validate_dataframe(df)
        if(ColumnNames.ACCELERATION not in df.columns):
            if(ColumnNames.VELOCITY not in df.columns):
                traces = velocity(df, per_traces=True)
            traces = acceleration(None, traces, per_traces=True)
        traces = jerkiness(None, traces, per_traces=True)

    return compute_metrics_from_traces(
        df=df,
        traces=traces,
        column_name=ColumnNames.JERKINESS,
        compute_traces_fn=lambda _: traces,  # Already computed above
        preprocess_fn=lambda s: s[s != 0]    # Exclude zero jerkiness
    )

def path(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None) -> pd.DataFrame:
    """
    Calculate the path length for the given DataFrame.
    This function computes the path length based on the Euclidean distance between consecutive points.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'x' and 'y' columns.
        traces (dict): A dictionary with keys as (sessionId) and values as lists of DataFrames. If None, traces will be computed from df.

    Returns:
        pd.DataFrame: DataFrame with an additional 'distance' column representing the path length.
    """

    if traces is None:
        validate_dataframe(df)
        traces = extract_traces_by_session(df)

    for session_id, session_traces in traces.items():
            for i in range(len(session_traces)):
                validate_dataframe(session_traces[i])

            # Compute the distance for each trace
            for j in range(len(session_traces)):
                session_traces[j]['distance'] = _path(session_traces[j])['distance']
            
            # Store the traces with distance in the dictionary
            traces[session_id] = session_traces

    return traces


def auc(df: pd.DataFrame, validation: bool = True, computeTraces: bool = True) -> float | dict:
    """
    Calculate the Area Under the Curve (AUC) for the given DataFrame.
    
    Parameters:
        df (pd.DataFrame): DataFrame containing 'timeStamp' and 'y' columns.
        validation (bool): Whether to validate the DataFrame structure, by default True.
        computeTraces (bool): Whether to compute traces by sessionId, by default True.
    
    Returns:
        float: The computed AUC value.
    """

    if(validation):
        validate_dataframe(df)

    # TODO revisar
    if computeTraces:
        df = df.sort_values(by=ColumnNames.TIME_STAMP)
        traces = extract_traces_by_session(df)
        auc_by_session = {}
        for session_id, session_traces in traces.items():
            session_traces = compute_space_time_diff(session_traces)
            auc_by_session[session_id] = np.mean([np.trapezoid(trace[ColumnNames.Y], trace[ColumnNames.X]) for trace in session_traces])
        return auc_by_session
    else:
        df = compute_space_time_diff(df)
        # Área bajo la curva real
        area_real = np.trapezoid(df[ColumnNames.Y], df[ColumnNames.X])
        return area_real


def auc_optimal(df: pd.DataFrame, validation: bool = True, computeTraces: bool = True) -> float:
    """
    Calculate the Optimal Area Under the Curve (AUC) for the given DataFrame.
    
    Parameters:
    df (pd.DataFrame): DataFrame containing 'timeStamp' and 'y' columns.
    validation (bool): Whether to validate the DataFrame structure, by default True.
    computeTraces (bool): Whether to compute traces by sessionId, by default True.
    
    Returns:
    float: The computed optimal AUC value.
    """
    
    if(validation):
        validate_dataframe(df)

    if computeTraces:
        df = extract_traces_by_session(df)


    df = compute_space_time_diff(df)

    # Área bajo la línea óptima
    x0, y0 = df[ColumnNames.X].iloc[0], df[ColumnNames.Y].iloc[0]
    x1, y1 = df[ColumnNames.X].iloc[-1], df[ColumnNames.Y].iloc[-1]
    x_opt = np.linspace(x0, x1, len(df))
    y_opt = np.linspace(y0, y1, len(df))
    area_optimal = np.trapezoid(y_opt, x_opt)

    return area_optimal

def auc_ratio(df: pd.DataFrame, traces: dict[str, list[pd.DataFrame]] = None, per_traces: bool = True) -> dict:
    """
    Calculate the AUC ratio for the given DataFrame.
    
    Parameters:
        df (pd.DataFrame): DataFrame containing 'timeStamp' and 'y' columns.
        computeTraces (bool): Whether to compute traces by sessionId, by default True. If False, df is assumed to be already segmented by sessionId.
    
    Returns:
        auc_per_session (dict): A dictionary with sessionId as keys and a tuple (auc, auc_ratio) as values. Where the auc is the area under the curve and auc_ratio is the ratio between the AUC and the optimal AUC.
    """

    if df is None and traces is None:
        raise ValueError("Either 'df' or 'traces' must be provided.")

    if not per_traces:
        # Compute directly on the DataFrame (no trace extraction)
        return auc_ratio_df(df)

    # If traces are not provided, extract them from df
    if traces is None:
        validate_dataframe(df)
        traces = extract_traces_by_session(df)

    return auc_ratio_traces(traces)

def auc_ratio_metrics(df: pd.DataFrame = None, computed_auc: dict = None, traces: dict[str, list[pd.DataFrame]] = None) -> dict:
    """
    This function computes the mean, max, and min auc ratio for each session on the given dataframe or list of traces.
    The optimal auc is not given in the output, only the auc ratio and auc, since it can be derived from them if needed.

    Parameters:
        df (pd.DataFrame): DataFrame containing interaction data.
        computed_auc (dict): Precomputed AUC ratios. If None, they will be computed from df or traces.
        traces (dict): A dictionary with keys as (sessionId) and values as lists of DataFrames. If None, traces will be computed from df.
    Returns:
        dict: A dictionary with keys as (sessionId) and values as dictionaries with 'mean_ratio', 'max_ratio', 'min_ratio' for the auc ratio and 'mean', 'max', and 'min' auc.
    """
    if (computed_auc is None):
        if(df is None):
            raise ValueError("Either 'df' or 'traces' must be provided.")
        if (traces is not None):
            computed_auc = auc_ratio(None, traces=traces, per_traces=True)
        else:
            computed_auc = auc_ratio(df)

    for session_id in computed_auc.keys():
        auc_ratios = [trace_metrics['auc_ratio'] for trace_metrics in computed_auc[session_id]]
        auc = [trace_metrics['auc'] for trace_metrics in computed_auc[session_id]]
        computed_auc[session_id] = {
            'mean_ratio': np.mean(auc_ratios) if auc_ratios else 0,
            'max_ratio': np.max(auc_ratios) if auc_ratios else 0,
            'min_ratio': np.min(auc_ratios) if auc_ratios else 0,
            'mean': np.mean(auc) if auc else 0,
            'max': np.max(auc) if auc else 0,
            'min': np.min(auc) if auc else 0,
        }
    return computed_auc
    

def deviations(df: pd.DataFrame = None, traces: dict[str, list[pd.DataFrame]] = None) -> dict:
    """
    Calculate the Mean Absolute Deviation (MAD) for the given DataFrame.

    Parameters:
        df (pd.DataFrame): DataFrame containing 'y' column.
        traces (dict): A dictionary with keys as (sessionId) and values as lists of DataFrames. If None, traces will be computed from df.
    Returns:
        dict: A dictionary with keys as (sessionId) and values as dictionaries with 'mad_mean' (mean of maximum absolute deviations), 'mad_max' (maximum absolute deviation across all traces), 'mad_min' (minimum absolute deviation across all traces) and 'aad' (average absolute deviation).
    """
    if traces is None:
        validate_dataframe(df)
        traces = extract_traces_by_session(df)

    metrics = {}
    for session_id, session_traces in traces.items():
        session_average_absolute_deviation = []
        session_mad_max = []
        for trace in session_traces:
            session_average_absolute_deviation.append(np.mean(np.abs(trace[ColumnNames.Y] - trace[ColumnNames.Y].mean())))
            session_mad_max.append(np.max(np.abs(trace[ColumnNames.Y] - trace[ColumnNames.Y].mean())))
        metrics[session_id] = {
            'aad': np.mean(session_average_absolute_deviation) if session_average_absolute_deviation else 0,
            'mad_max': np.max(session_mad_max) if session_mad_max else 0,
            'mad_mean': np.mean(session_mad_max) if session_mad_max else 0,
            'mad_min': np.min(session_mad_max) if session_mad_max else 0,
        }

    return metrics


import matplotlib.pyplot as plt
import pandas as pd
import numpy as np
import cv2
from pywib.constants import EventTypes

from pywib.constants import ColumnNames

def visualize_trace(df, stroke_indices, stroke_id):
    """
    Generates a plot visualizing the trace of a stroke.

    Parameters:
        df (pd.DataFrame): DataFrame containing the stroke data with 'x', 'y and 'timeStamp' columns.
        stroke_indices (list): List of indices corresponding to the stroke in the DataFrame. Can be obtained using df.index.
        stroke_id (str): Identifier for the stroke to be displayed in the title.
    Returns:
        None: Displays a plot of the stroke trace.
    """
    stroke_data = df.loc[stroke_indices]
    plt.figure(figsize=(10, 8))
    plt.plot(stroke_data['x'], stroke_data['y'], 'b-o', linewidth=2, markersize=4, label='Trazo real')

    x_start, y_start = stroke_data['x'].iloc[0], stroke_data['y'].iloc[0]
    x_end, y_end = stroke_data['x'].iloc[-1], stroke_data['y'].iloc[-1]
    plt.plot([x_start, x_end], [y_start, y_end], 'r--', linewidth=2, label='Línea óptima')

    plt.plot(x_start, y_start, 'go', markersize=8, label='Inicio')
    plt.plot(x_end, y_end, 'ro', markersize=8, label='Fin')

    duration = stroke_data['timeStamp'].iloc[-1] - stroke_data['timeStamp'].iloc[0]
    plt.xlabel('X (píxeles)')
    plt.ylabel('Y (píxeles)')
    plt.title(f'Trazo {stroke_id} - Duración: {duration:.0f}ms - Puntos: {len(stroke_data)}')
    plt.legend()
    plt.grid(True, alpha=0.3)
    plt.gca().invert_yaxis()
    plt.show()

def video_from_trace(df, user_id, outfile: str, width=640, height=480, fps=30, colored=False):
    """
    Generates a video visualizing the mouse trace of a user.
    
    Parameters:
        df (pd.DataFrame): DataFrame containing the mouse trace data with 'x', 'y', 'eventType', and 'sessionId' columns.
        user_id (str/int): Identifier of the user whose trace is to be visualized.
        outfile (str): Path to save the output video file.
        width (int): Width of the video frame.
        height (int): Height of the video frame.
        fps (int): Frames per second for the video.
        colored (bool): If True, applies a temperature gradient (green to red) to the trajectory line.
    Returns:
        None: Saves the video file to the specified path.
    """
    user_data = df[df[ColumnNames.SESSION_ID] == user_id].sort_values(ColumnNames.TIME_STAMP)
    xs = user_data[ColumnNames.X].values
    ys = user_data[ColumnNames.Y].values
    eventType = user_data[ColumnNames.EVENT_TYPE].values
    n = len(xs)
    # Nombre del vídeo
    fourcc = cv2.VideoWriter_fourcc(*'mp4v')
    video = cv2.VideoWriter(outfile, fourcc, fps, (width, height))
    
    # Generar frames
    frame = np.ones((height, width, 3), dtype=np.uint8) * 255  # fondo blanco
    last_x, last_y = xs[0], ys[0]
    for i in range(1, n):
        # Calculate color based on progress if colored mode is enabled
        if colored:
            # Temperature gradient: green (0, 255, 0) -> yellow -> red (0, 0, 255)
            progress = i / (n - 1)
            green = int(255 * (1 - progress))
            red = int(255 * progress)
            line_color = (0, green, red)  # BGR format
        else:
            line_color = (0, 0, 255)  # Default red
        
        # Dibujar línea de movimiento
        if xs[i] <= 0 or ys[i] <= 0 or xs[i] > width or ys[i] > height:
            # Coordenadas fuera de la pantalla, no dibujar
            pass
        elif xs[i-1] <= 0 or ys[i-1] <= 0 or xs[i-1] > width or ys[i-1] > height:
            # Coordenadas anteriores fuera de la pantalla
            if(last_x > 0 and last_y > 0):
                cv2.line(frame, (last_x, last_y), (xs[i], ys[i]), color=line_color, thickness=2)
        else:
            cv2.line(frame, (xs[i-1], ys[i-1]), (xs[i], ys[i]), color=line_color, thickness=2)
        
        # Actualizar última posición válida a la posición actual si es válida (dentro de la pantalla)
        if xs[i] > 0 and ys[i] > 0 and xs[i] <= width and ys[i] <= height:
            last_x, last_y = xs[i], ys[i]

        # Dibujar clics
        if eventType[i] == EventTypes.EVENT_ON_CLICK or (eventType[i-1] == EventTypes.EVENT_ON_MOUSE_DOWN and eventType[i] == EventTypes.EVENT_ON_MOUSE_UP):
            cv2.circle(frame, (last_x, last_y), radius=5, color=(0, 0, 255), thickness=-1)
        elif eventType[i] == EventTypes.EVENT_KEY_DOWN or eventType[i] == EventTypes.EVENT_KEY_UP:
            # This event has coordinates (0,0) or (-1,-1)
            cv2.rectangle(frame, (last_x-5, last_y-5), (last_x+5, last_y+5), color=(255, 0, 0), thickness=-1)
        elif eventType[i] == EventTypes.EVENT_WINDOW_SCROLL:
            # This event has coordinates (0,0) or (-1,-1)
            pts = np.array([(last_x, last_y-5), (last_x-5, last_y+5), (last_x+5, last_y+5)], np.int32)
            cv2.polylines(frame, [pts], True, (255, 0, 0), 2)
        else:
            # Unknown or unhandled event type
            pass
        
        # Escribir frame en vídeo
        video.write(frame.copy())
    
    video.release()
    print(f"Vídeo generado para el usuario {user_id}: {outfile}")


"""
"""

__version__ = "0.1.0"
__author__ = "Guillermo Dylan Carvajal Aza"
__email__ = "carvajalguillermo@uniovi.es"

from .constants import *
from .utils import (validate_dataframe, validate_dataframe_keyboard, 
                    extract_traces_by_session, visualize_trace, compute_space_time_diff, video_from_trace, validate_duplicate_timestamps)
from .core import (velocity, acceleration, jerkiness, path, auc_ratio, 
                   execution_time, movement_time, pauses_metrics, velocity_metrics, 
                   acceleration_metrics, jerkiness_metrics, number_of_clicks, 
                   click_slip, num_pauses, deviations, auc_ratio_metrics, typing_speed_metrics, typing_speed, backspace_usage)

__all__ = [
    # Version info
    "__version__",
    "__author__", 
    "__email__",

    # Constants
    "EventTypes",
    "ComponentTypes",
    
    # Utility functions
    "validate_dataframe",
    "validate_dataframe_keyboard",
    "extract_trace",
    "visualize_trace",
    "compute_space_time_diff",
    "extract_traces_by_session",
    "video_from_trace",
    "validate_duplicate_timestamps",

    # Movement functions
    "velocity",
    "acceleration",
    "auc_ratio",
    "jerkiness",
    "path",
    "velocity_metrics",
    "acceleration_metrics",
    "jerkiness_metrics",
    "deviations",
    "auc_ratio_metrics",

    # Mouse functions
    "number_of_clicks",
    "click_slip",

    # Keyboard functions
    "typing_speed",
    "typing_speed_metrics",
    "backspace_usage",

    # Timing
    "pauses_metrics",
    "execution_time",
    "movement_time",
    "num_pauses",

]


import unittest
import numpy as np
import sys
import os
import pytest
sys.path.insert(0, os.path.dirname(__file__))
from utils import process_csv, import_pyModule

import_pyModule()

from pywib import typing_speed_metrics, typing_speed, backspace_usage

DEBUG = True

class TestData:
    if(DEBUG):
        dataFile = 'test/test_data/test_mouse_keyboard.csv'
    else:
        dataFile = 'pywib/test/test_data/test_mouse_keyboard.csv'

    reults = {
        "SESSION_A":{
            "speed": [float((1/((1750792681050 - 1750792680900)/1000.0)) * 60.0), float(1/((1750792685500 - 1750792685400)/1000.0) * 60.0)],
            "backspace_usage": 0,
            "typing_durations": [120, 130, 115, 125, 135],
            "total_chars": 2
        },
        "SESSION_B":{

            "speed": [float( (1/((1750792680980 - 1750792680920)/1000.0)) * 60.0), float( (4/((1750792686150 - 1750792685600)/1000.0)) * 60.0)],
            "backspace_usage": 3,
            "typing_durations": [100, 110, 90, 95, 105],
            "total_chars": 5
        }
    }

class TestKeyboard(unittest.TestCase):
    
    def setUp(self):
        """Set up test data"""
        # Create sample test data instead of relying on external CSV
        self.test_data = process_csv(TestData.dataFile)
        
    def test_typing_speed(self):
        """Test typing_speed function"""
        speed = typing_speed(self.test_data, per_traces=True)
        for session_id, speeds in speed.items():
            expected_speeds = TestData.reults[session_id]["speed"]
            for s, e in zip(speeds, expected_speeds):
                self.assertAlmostEqual(s, e, places=2)

    def test_typing_speed_df(self):
        """Test typing_speed function for DataFrame input"""
        df_session_b = self.test_data[self.test_data['sessionId'] == 'SESSION_B']
        speed = typing_speed(df_session_b, per_traces=False)
        self.assertEqual(speed, (6/(2500/1000)) * 60.0)

    def test_backspace_usage(self):
        """Test backspace_usage function"""
        backspace_stats = backspace_usage(self.test_data)
        for session_id, usage in backspace_stats.items():
            self.assertEqual(usage, TestData.reults[session_id]["backspace_usage"])


    def test_typing_speed_df(self):
        """Test typing_speed function for DataFrame input"""
        df_session_b = self.test_data[self.test_data['sessionId'] == 'SESSION_B']
        speed = typing_speed(df_session_b, per_traces=False)
        self.assertEqual(speed, (5/((1750792686150 - 1750792680920)/1000.0)) * 60.0)

    def test_typing_speed_metrics(self):
        """Test typing_speed_metrics function"""
        metrics = typing_speed_metrics(self.test_data)
        print(metrics)
        for session_id, metric in metrics.items():
            expected_speeds = TestData.reults[session_id]["speed"]
            avg_expected_speed = sum(expected_speeds) / len(expected_speeds)
            self.assertAlmostEqual(metric["average_typing_speed"], avg_expected_speed, places=2)
            self.assertEqual(metric["total_characters"], TestData.reults[session_id]["total_chars"])

if __name__ == '__main__':
    unittest.main()